/*
 * PairwiseAlignment.cpp
 *
 *  Created on: Apr 19, 2010
 *      Author: Oleksiy Kohany <kohany@gmail.com>
 */

#include "PairwiseAlignment.h"
#include <iomanip>

namespace bio {

   vector<string> PairwiseAlignment::ignoreMasks;
   PairwiseAlignment::ScoreMatrix_t PairwiseAlignment::scoreMatrix;
   string PairwiseAlignment::scoreMatrixName;

   PairwiseAlignment::PairwiseAlignment() :
      score(0) {
   }
   PairwiseAlignment::PairwiseAlignment(const PairwiseAlignment& other) :
      top(other.getTop()), bot(other.getBot()), score(other.getScore()) {
   }
   PairwiseAlignment::PairwiseAlignment(const NamedSubSequence& top, const NamedSubSequence& bot, score_t score) :
      top(top), bot(bot), score(score) {
   }
   PairwiseAlignment::~PairwiseAlignment() {
   }
   void PairwiseAlignment::printAlnHead(ostream& os) const {
      PairwiseAlignmentStats stat;
      accumulateAlignmentStats(top.begin(), top.end(), bot.begin(), bot.end(), ignoreMasks.begin(),
                                                               ignoreMasks.end(), scoreMatrix, stat);
      //m_sCon.guessType();
      //m_sSeq.guessType();
      os << setw(12) << setiosflags(ios::left) << top.getName() << " ";
      os << setw(7) << resetiosflags(ios::left) << top.getFragments().front().getBeg() + 1 << " ";
      os << setw(7) << setiosflags(ios::left) << top.getFragments().front().getEnd() << " ";
      os << setw(12) << resetiosflags(ios::left) << bot.getName() << " ";
      os << setw(7) << bot.getFragments().front().getBeg() + 1 << " ";
      os << setw(7) << setiosflags(ios::left) << bot.getFragments().front().getEnd();
      os << ((top.isDirect() ^ bot.isDirect()) ? " c  " : " d  ");
      os << " " << setw(5) << setiosflags(ios::left) << setprecision(4) << stat.getSim();
      os << " " << setw(5) << setprecision(2) << ((scoreMatrix.getType() == NA) ? stat.getTrs() : stat.getPos());
      os << " " << setw(5) << resetiosflags(ios::left) << (int) getScore();
   }
   void PairwiseAlignment::printAlnBody(ostream& os, size_t width, const char* sBlahS, const char* sBlahQ) const {
      string sCon;
      string sSeq;
      string::iterator cPtr, sPtr;
      for (size_t i = 0; i < top.size(); i += width) {
         (os << "S: ").write(top.c_str() + i, min(width, top.size() - i)) << (sBlahS?sBlahS:"") << endl <<"   ";
         for (size_t j = i; j < top.size() && j < i + width; j++) {
            os << getLink(top[j], bot[j], scoreMatrix);
         }
         (os << "\nQ: ").write(bot.c_str() + i, min(width, top.size() - i)) << (sBlahQ?sBlahQ:"") << "\n" << endl;
      }
   }
   void PairwiseAlignment::setScoreMatrix(const ScoreMatrix_t& matrix) {
      scoreMatrix = matrix;
   }
   void PairwiseAlignment::loadScoreMatrix(const char* s) {
      if (scoreMatrixName.compare(s) != 0) {
         if (scoreMatrix.parse(s)) {
            scoreMatrixName = s;
         }
      }
   }
   PairwiseAlignmentStats PairwiseAlignment::getStat() {
      PairwiseAlignmentStats stat;
      return accumulateAlignmentStats(top.begin(), top.end(), bot.begin(), bot.end(), ignoreMasks.begin(), ignoreMasks.end(), scoreMatrix, stat);
   }


   bool isTransition(char c1, char c2) {
      //                                        A, B, C, D, E, F, G, H, I, J, K, L, M, N, O,
      //                                     P, Q, R, S, T, U, V, W, X, Y, Z, [, \, ], ^, _,
      //                                     `, a, b, c, d, e, f, g, h, i, j, k, l, m, n, o,
      //                                     p, q, r, s, t, u, v, w, x, y, z
      static const unsigned char index[] = { 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
                                             4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
                                             4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
                                             4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
                                             4, 0, 4, 1, 4, 4, 4, 2, 4, 4, 4, 4, 4, 4, 4, 4,
                                             4, 4, 4, 4, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
                                             4, 0, 4, 1, 4, 4, 4, 2, 4, 4, 4, 4, 4, 4, 4, 4,
                                             4, 4, 4, 4, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
                                             4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
                                             4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
                                             4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
                                             4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
                                             4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
                                             4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
                                             4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
                                             4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4 };
      //                                 A  C  G  T
      static const bool trans[5][5] = {{ 0, 0, 1, 0, 0 }, //A
                                       { 0, 0, 0, 1, 0 }, //C
                                       { 1, 0, 0, 0, 0 }, //G
                                       { 0, 1, 0, 0, 0 }, //T
                                       { 0, 0, 0, 0, 0 } };
      {
         return trans[index[c1]][index[c2]];
      }
      return false;
   }
   char getLink(char c1, char c2, PairwiseAlignment::ScoreMatrix_t& matrix) {
      if(matrix.getChar2Index()[c1] == matrix.getChar2Index()[c2]) {
         if(matrix.getScore(c1, c2) > 0) {
            return '|';
         }
         return ' ';
      }
      if(matrix.getScore(c1, c2) > 0) {
         return '+';
      }
      if(matrix.getType() == NA && isTransition(c1, c2)) {
         return ':';
      }
      if(c1 == matrix.GAPCHAR || c2 == matrix.GAPCHAR) {
         return '-';
      }
      return ' ';
   }

   PairwiseAlignment& sharpen(PairwiseAlignment& aln, size_t edgeWidht) {
      if (edgeWidht <= 0) {
         return aln;
      }
      int aln1 = -1;
      int aln2 = -1;
      int edge = 0;
      int start = 0;
      for (int i = 0; i < aln.getBot().size(); i++) {
         if (aln.getBot()[i] == aln.getTop()[i]) {
            if (isalpha(aln.getBot()[i]) || aln.getBot()[i] == aln.getScoreMatrix().STOPCODON) {
               edge++;
            }
         } else {
            edge = 0;
            start = i + 1;
         }
         if (edge >= edgeWidht) {
            aln1 = start;
            break;
         }
      }
      trimLeft(aln, aln1);

      edge = 0;
      start = aln.getBot().size() - 1;
      for (int i = start; i >= 0; i--) {
         if (aln.getBot()[i] == aln.getTop()[i]) {
            if (isalpha(aln.getBot()[i]) || aln.getBot()[i] == aln.getScoreMatrix().STOPCODON) {
               edge++;
            }
         } else {
            edge = 0;
            start = i - 1;
         }
         if (edge >= edgeWidht) {
            aln2 = start;
            break;
         }
      }
      trimRight(aln, aln.getBot().size() - aln2 - 1);
      return aln;
   }

   PairwiseAlignment& trimLeft(PairwiseAlignment& aln, size_t trimLength) {
      if (trimLength >= aln.getBot().size()) {
         aln.setScore(0);
         aln.getBot().clear();
         aln.getTop().clear();
         return aln;
      }
      int refDelta = aln.getBot().getBeg() <= aln.getBot().getEnd() ? 1 : -1;
      int seqDelta = aln.getTop().getBeg() <= aln.getTop().getEnd() ? 1 : -1;

      for (size_t i = 0; i < trimLength && i < aln.getBot().size(); i++) {
         if (isalpha(aln.getBot()[i]) || aln.getBot()[i] == aln.getScoreMatrix().STOPCODON) {
            aln.getBot().setBeg(aln.getBot().getBeg() + refDelta);
         }
         if (isalpha(aln.getTop()[i]) || aln.getTop()[i] == aln.getScoreMatrix().STOPCODON) {
            aln.getTop().setBeg(aln.getTop().getBeg() + seqDelta);
         }
      }
      aln.setScore(accumulateScore(aln.getTop().begin() + trimLength, aln.getTop().end(), aln.getBot().begin() + trimLength, aln.getBot().end(), (double*)NULL, (double*)NULL, aln.getScoreMatrix()));
      aln.getBot().erase(0, trimLength);
      aln.getTop().erase(0, trimLength);
      return aln;
   }

   PairwiseAlignment& trimRight(PairwiseAlignment& aln, size_t trimLength) {
      if (trimLength >= aln.getBot().size()) {
         aln.setScore(0);
         aln.getBot().clear();
         aln.getTop().clear();
         return aln;
      }
      int refDelta = aln.getBot().getBeg() <= aln.getBot().getEnd() ? 1 : -1;
      int seqDelta = aln.getTop().getBeg() <= aln.getTop().getEnd() ? 1 : -1;

      for (size_t x = 0; x < trimLength && x < aln.getBot().size(); x++) {
         int i = aln.getBot().size() - 1 - x;
         if (isalpha(aln.getBot()[i]) || aln.getBot()[i] == aln.getScoreMatrix().STOPCODON) {
            aln.getBot().setEnd(aln.getBot().getEnd() - refDelta);
         }
         if (isalpha(aln.getTop()[i]) || aln.getTop()[i] == aln.getScoreMatrix().STOPCODON) {
            aln.getTop().setEnd(aln.getTop().getEnd() - seqDelta);
         }
      }
      aln.setScore(accumulateScore(aln.getTop().begin(), aln.getTop().end() - trimLength,
                                   aln.getBot().begin(), aln.getBot().end() - trimLength,
                                   (double*)NULL, (double*)NULL, aln.getScoreMatrix()));

      aln.getBot().erase(aln.getBot().size() - trimLength, trimLength);
      aln.getTop().erase(aln.getTop().size() - trimLength, trimLength);
      return aln;
   }

   istream& operator>>(istream& is, PairwiseAlignment& aln) {

   }
   ostream& operator<<(ostream& os, const PairwiseAlignment& aln) {
      aln.printAlnHead(os);
      os << " <\n\n";
      aln.printAlnBody(os);
      return os;
   }
   ostream& operator<<(ostream& os, const PairwiseAlignmentStats& v) {
      os << "m= " << v.iMCnt << "\tmm= " << (v.iMMCnt - v.iGCnt) << "\tts= " << v.iTRCnt << "\tgs= " << v.iGCnt << "\ts= " << ((v.iMCnt + v.iMMCnt
            == 0) ? -1.0 : (double) v.iMCnt / (v.iMCnt + v.iMMCnt)) << "\tts/tv= " << ((v.iTVCnt == 0) ? -v.iTRCnt : (double) v.iTRCnt / v.iTVCnt)
            << " #\n";
   }

}
