/*
 * Sequence.h
 *
 *  Created on: Apr 15, 2010
 *      Author: Oleksiy Kohany <kohany@gmail.com>
 */

#ifndef SEQUENCE_H_
#define SEQUENCE_H_

#include <string>
#include <algorithm>
#include <ctype.h>
#include <iostream>
#include <stdexcept>
using namespace std;
/*
 *
 */
namespace bio {
   class Sequence {
      string data;
      virtual void readFAHeaderLine(const string& s);
      virtual void readIGHeaderLine(const string& s);
   public:
      static const char GAPCHAR;
      typedef char char_t;
      typedef string::iterator iterator;
      typedef string::const_iterator const_iterator;
      typedef string::reverse_iterator reverse_iterator;
      typedef string::const_reverse_iterator const_reverse_iterator;
      Sequence();
      Sequence(const Sequence& other);
      Sequence(const string& data);
      Sequence(const char* data);
      virtual ~Sequence();
      Sequence& operator=(const string& other) {
         data = other;
         return *this;
      }
      Sequence& operator=(const char* other) {
         data = other;
         return *this;
      }
      Sequence& operator=(const Sequence& other) {
         data=(other.data);
         return *this;
      }
      Sequence& operator+=(char c) {
         data+=c;
         return *this;
      }
      bool operator==(const Sequence& other) {
         return data==other.data;
      }
      iterator begin() {
         return data.begin();
      }
      iterator end() {
         return data.end();
      }
      const_iterator begin() const {
         return data.begin();
      }
      const_iterator end() const {
         return data.end();
      }
      reverse_iterator rbegin() {
         return data.rbegin();
      }
      reverse_iterator rend() {
         return data.rend();
      }
      const_reverse_iterator rbegin() const {
         return data.rbegin();
      }
      const_reverse_iterator rend() const {
         return data.rend();
      }
      char_t& operator[](size_t pos) {
         return data[pos];
      }
      const char_t& operator[](size_t pos) const {
         return data[pos];
      }
      void push_back ( char_t c ) {
         data.push_back(c);
      }
      size_t size() const {
         return data.size();
      }
      const char_t* c_str ( ) const {
         return data.c_str();
      }
      virtual void clear() {
         data.clear();
      }
      Sequence& clean();
      void resize(size_t size) {
         data.resize(size);
      }
      Sequence& remove(char_t c);
      operator string&() {
         return data;
      }
      operator const string&() const {
         return data;
      }
      void toUpper() {
         for_each(data.begin(), data.end(), ::toupper);
      }
      void toLower() {
         for_each(data.begin(), data.end(), ::tolower);
      }
      Sequence& insert ( size_t pos1, size_t n, char c ) {
         data.insert(pos1, n, c);
         return *this;
      }
      iterator insert ( iterator p, char c ) {
         return data.insert(p, c);
      }
      void insert ( iterator p, size_t n, char c ) {
         data.insert(p, n, c);
      }
      Sequence& erase ( size_t pos = 0, size_t n = string::npos ) {
         data.erase(pos, n);
         return *this;
      }
      Sequence& read(istream& is);
      bool readHeader(istream& is);
      bool readFAHeader(istream& is);
      bool readIGHeader(istream& is);
      void print(ostream& os, bool format=true) const;
      void printFA(ostream& os, bool format=true) const;
      void printIG(ostream& os, bool format=true) const;
      void printHeader(ostream& os) const;
      virtual void printFAHeader(ostream& os, const char* comment = NULL) const;
      virtual void printIGHeader(ostream& os, const char* comment = NULL) const;
      void printFormattedBody(ostream& os, size_t width = 70) const;
      void printUnformattedBody(ostream& os) const;
   };
   string& clean(string& s);
   char complement(char c);
   ostream& operator<<(ostream& os, const Sequence& seq);
   istream& operator>>(istream& is, Sequence& seq);

   template <typename InputIterator>
   size_t bioLength(InputIterator beg, InputIterator end) {
      size_t length = 0;
      for(InputIterator ptr = firstBase(beg, end); ptr != end; ptr = firstBase(++ptr, end)) {
         ++length;
      }
      return length;
   }

   template <typename ForwardIterator>
   ForwardIterator firstBase(ForwardIterator ptr, ForwardIterator end) {
      //                          "*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz"
      static const char index[] = "100000000000000000000001111111111111111111111111100000011111111111111111111111111";
      while(ptr != end) {
         if (*ptr >= '*' && *ptr <= 'z' && index[*ptr - '*']) {
            break;
         }
         ++ptr;
      }
      return ptr;
   }

   /* Reverse complements sequence of nucleotide symbols in the range beg-end. */
   template <typename RandomAccessIterator>
   void complement(RandomAccessIterator beg, RandomAccessIterator end) {
      if (beg < end) {
         --end;
         while (beg <= end) {
            typename iterator_traits<RandomAccessIterator>::value_type c = *beg;
            *beg = complement(*end);
            *end = complement(c);
            ++beg;
            --end;
         }
      }
      else {
         throw invalid_argument("ERROR: Expecting valid range 'beg < end'");
      }
   }
}

#endif /* SEQUENCE_H_ */
